from pathlib import Path
import shutil

from grid.downloader import DownloadableObject
from grid.downloader import Downloader


class TestArtifactCallbacks:
    """Tests callbacks in grid artifacts."""
    @classmethod
    def clean_test_directory(cls):
        path = Path(cls.test_dir)
        shutil.rmtree(path, ignore_errors=True)

    @classmethod
    def setup_class(cls):
        cls.test_dir = 'tests/data/download_path/nested_dir'
        cls.clean_test_directory()

    @classmethod
    def teardown_class(cls):
        # Deletes test directory with all files.
        cls.clean_test_directory()

    def test_downloader_creates_dir(self):
        """Downloader.create_dir_tree creates dir structure in path."""
        Downloader.create_dir_tree(self.test_dir)
        assert Path(self.test_dir).exists()

    def test_downloader_downloads_data(self):
        """Downloader().download() downloads data in nested directories."""
        # Google logo
        # skipcq: FLK-E501
        url = 'https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_272x92dp.png'
        directory_path = 'nested_0/nested_1'
        filename = 'test-filename.png'
        objects = [
            DownloadableObject(url=url,
                               download_path=directory_path,
                               filename=filename)
        ]

        # Download file
        D = Downloader(downloadable_objects=objects, base_dir=self.test_dir)
        assert D.download()

        # Test that file as been downloaded.
        assert Path(self.test_dir).joinpath(directory_path).joinpath(
            filename).exists()



# Copyright 2020 Grid AI Inc.
from array import array
import ast
import base64
import csv
from datetime import datetime
import json
import os
from pathlib import Path
import time
from typing import Any, Callable, Dict, List, Optional
from urllib.parse import urljoin
from urllib.parse import urlsplit
import zlib

import click
from dateutil.parser import parse as date_string_parse
from gql import Client
from gql import gql
from gql.transport.exceptions import TransportProtocolError
from gql.transport.exceptions import TransportQueryError
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.websockets import WebsocketsTransport
import humanize
from packaging import version
import requests
from requests.exceptions import HTTPError
from rich.console import Console
import websockets
import yaml
import yaspin

from grid.commands import CredentialsMixin
from grid.commands import DependencyMixin
from grid.commands import WorkflowChecksMixin
from grid.commands.git import execute_git_command
from grid.datastore import create_datastore_session
from grid.datastore import DatastoreUploadSession
from grid.downloader import DownloadableObject
from grid.downloader import Downloader
from grid.exceptions import AuthenticationError
from grid.exceptions import TrainError
import grid.globals as env
from grid.metadata import __version__
from grid.observables import BaseObservable
from grid.observables import Experiment
from grid.observables import InteractiveNode
from grid.observables import Run
from grid.types import ObservableType
from grid.types import WorkflowType


class Grid(CredentialsMixin, WorkflowChecksMixin, DependencyMixin):
    """
    Interface to the Grid API.

    Attributes
    ----------
    url: str
        Grid URL
    request_timeout: int
        Number of seconds to timeout a request by default.
    client: Client
        gql client object
    grid_credentials_path: str
        Path to the Grid credentials
    default_headers: Dict[str, str]
        Header used in the request made to Grid.
    acceptable_lines_to_print: int
        Total number of acceptable lines to print in
        stdout.
    request_cooldown_duration: float
        Number of seconds to wait between continuous
        requests.

    Parameters
    ----------
    local_credentials: bool, default True
        If the client should be initialized with
        credentials from a local file or not.
    """
    url: str = env.GRID_URL

    #  TODO: Figure out a better timeout based on query type.
    request_timeout: int = 60
    default_headers: Dict[str, str] = {
        'Content-type': 'application/json',
        'User-Agent': f'grid-api-{__version__}'
    }

    grid_settings_path: str = '.grid/settings.json'
    grid_credentials_path: str = '.grid/credentials.json'

    client: Client
    transport: RequestsHTTPTransport

    available_observables: Dict[ObservableType, Callable] = {
        ObservableType.EXPERIMENT: Experiment,
        ObservableType.RUN: Run,
        ObservableType.INTERACTIVE: InteractiveNode
    }

    acceptable_lines_to_print: int = 50
    request_cooldown_duration: int = 0.1

    def __init__(self,
                 credential_path: Optional[str] = None,
                 load_local_credentials: bool = True):

        self.credentials: Dict[str, str] = {}
        self.credential_path = credential_path
        self.headers = self.default_headers.copy()

        #  By default, we instantiate the client with a local
        #  set of credentials.
        if load_local_credentials or self.credential_path:
            self._set_local_credentials()

            #  The client will be created with a set of credentials.
            #  If we change these credentials in the context of a
            #  call, for instance "login()" then we have to
            #  re-instantiate these credentials.
            self._init_client()

        # Loads global settings on startup.
        # Also creates settings if they are not
        # available.
        self._load_global_settings()
        super().__init__()

    @property
    def user_id(self):
        return self.credentials.get('UserID')

    def _set_local_credentials(self):
        """
        Instantiates the GraphQL local client using local credentials.
        """
        #  Re-fetches values from env.
        env.USER_ID = os.getenv('GRID_USER_ID')
        env.API_KEY = os.getenv('GRID_API_KEY')
        if env.USER_ID and env.API_KEY:
            click.echo('Configuring user from environment')
            self.__set_authentication_headers(username=env.USER_ID,
                                              key=env.API_KEY)
            return

        #  Checks if the environment variable GRID_CREDENTIAL_PATH
        #  contains a path for grid credentials.
        #  TODO: Click has a better interface for doing this.
        env_path = os.getenv('GRID_CREDENTIAL_PATH')
        if env_path:
            P = Path(env_path)
            if not P.exists():
                m = f'Credentials not found at {env_path}. Did you set GRID_CREDENTIAL_PATH correctly?'
                raise click.ClickException(m)
        elif self.credential_path:
            P = Path(self.credential_path)
            if not P.exists():
                m = f'Credentials not found at {self.credential_path}'
                raise click.ClickException(m)
        else:
            P = Path.home().joinpath(self.grid_credentials_path)

        if P.exists():
            self.credentials = json.load(P.open())
            self.__set_authentication_headers(
                username=self.credentials['UserID'],
                key=self.credentials['APIKey'])

        else:
            raise click.ClickException(
                'No credentials available. Did you login?')

    def __set_authentication_headers(self, username: str, key: str) -> None:
        """
        Sets credentials header for a client.
        """
        self.headers['X-Grid-User'] = username
        self.headers['X-Grid-Key'] = key

    def _load_global_settings(self) -> None:
        """
        Loads user settings and sets them globally
        in the Client context.
        """
        P = Path.home().joinpath(self.grid_settings_path)

        # Make sure path exists.
        Path(P.parents[0]).mkdir(parents=True, exist_ok=True)

        # If file doesn't exist, create with default global
        # settings.
        if not P.exists():
            global_variables = {
                'debug': False,
                'ignore_warnings': False,
                'skip_vesion_check': False
            }
            with P.open('w') as file:
                json.dump(global_variables, file, ensure_ascii=False, indent=4)

        # Setup settings based on what the user has
        # configured.
        else:
            user_settings = json.load(P.open())
            if 'debug' in user_settings and env.DEBUG is None:
                env.DEBUG = bool(user_settings['debug'])

            if 'ignore_warnings' in user_settings and env.IGNORE_WARNINGS is None:
                env.IGNORE_WARNINGS = bool(user_settings['ignore_warnings'])

            if 'skip_version_check' in user_settings and env.SKIP_VERSION_CHECK is None:
                env.SKIP_VERSION_CHECK = bool(
                    user_settings['skip_vesion_check'])

    def _check_version_compatibility(self):
        """
        Checks API version compatibility between server and client.
        Raises exception if versions are incompatible.
        """
        # Skip version check if configured by user.
        if env.SKIP_VERSION_CHECK:
            return

        _url = self.url.replace('graphql', '')
        _url = urljoin(_url, 'metadata')
        response = requests.get(_url)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError:
            raise click.ClickException('Grid is currently unavailable. '
                                       'Try again in a few minutes.')

        data = response.json()
        server_version = data['version']

        # Compare versions.
        message = """You need to update Grid! Please Run:

    pip install lightning-grid --upgrade
"""
        grid_version = version.parse(server_version)
        client_version = version.parse(__version__)

        # Right now any version differences will
        # raise an exception. We can be a lot more detailed
        # in the future and only error on major differences.
        if grid_version > client_version:
            if grid_version.major > client_version.major:
                click.echo(message)
                raise click.ClickException(
                    'Incompatible versions: {grid_version} (server) and '
                    '{client_version} (client)')

    def _init_client(self, websocket: bool = False) -> None:
        """
        Initializes GraphQL client. This fetches the latest
        schema from Grid.
        """
        # Check version compatibility on client initialization.
        # TODO re-enable when versioning is fixed
        # self._check_version_compatibility()

        if websocket:
            _url = self.url.replace('http://', 'ws://')
            _url = _url.replace('https://', 'wss://')
            _url = _url.replace('graphql', 'subscriptions')
            self.transport = WebsocketsTransport(url=_url,
                                                 init_payload=self.headers)
        else:
            self.transport = RequestsHTTPTransport(
                url=self.url,
                use_json=True,
                headers=self.headers,
                timeout=self.request_timeout,
                retries=3)

        try:
            self.client = Client(transport=self.transport,
                                 fetch_schema_from_transport=True)

        except requests.exceptions.ConnectionError:
            raise click.ClickException(
                f'Grid is unreachable. Is Grid online at {env.GRID_URL.replace("/graphql", "")} ?'
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                raise click.ClickException('Not authorized. Did you login?')
            if e.response.status_code == 500:
                raise click.ClickException(
                    'Grid is having issues. Please again later.')
            raise click.ClickException('We encountered an unknown error.')

        except requests.exceptions.Timeout:
            raise click.ClickException('Could not reach Grid. Are you online?')

        except TransportProtocolError:
            raise click.ClickException('Not authorized. Did you login?')

        except Exception as e:
            raise click.ClickException(f"{type(e).__name__}: {e}")

    def execute_gql(self, query: str, **kwargs) -> Dict:
        values = kwargs or {}
        try:
            result = self.client.execute(gql(query), variable_values=values)
        except TransportQueryError as e:
            raise click.ClickException(e.args)
        except Exception as e:
            raise click.ClickException(f"{type(e).__name__}: {e}")
        if 'error' in result:
            raise click.ClickException(json.dumps(result['error']))
        return result

    @staticmethod
    def _add_git_root_path(entrypoint: str) -> str:
        #  Finds the relative path of the file to train.
        repository_path = execute_git_command(['rev-parse', '--show-toplevel'])
        current_path = str(Path.cwd())
        script_path = current_path.replace(str(repository_path), '')
        env.logger.debug(script_path)

        _entrypoint = str(Path(script_path).joinpath(entrypoint))
        return _entrypoint

    def _check_user_github_token(self) -> bool:
        """
        Checks if user has a valid Github token available.
        If user doesn't have one, then redirect user to the
        Grid UI to fetch a new one.

        Returns
        -------
        has_token: bool
            Boolean indicating if user has valid token.
        """
        # Build query
        query = gql("""
            query CheckToken {
                checkUserGithubToken {
                    hasValidToken
                }
            }
        """)

        # Check if the user has a token. If she hasn't,
        # then redirect user to the /auth page to get a new
        # Github token.
        has_token = False
        try:
            result = self.client.execute(query)
            has_token = result['checkUserGithubToken']['hasValidToken']
            if not has_token:
                auth_url = env.GRID_URL.replace("graphql", "auth")
                click.launch(auth_url)
                raise click.ClickException("""
    Authentication tokens need to be renewed! Opening Grid on the browser so
    we can renew your authentication tokens.
    """)

        except HTTPError as e:
            click.echo(str(e), err=True)
            raise AuthenticationError(e)

        return has_token

    def login(self, username: str, key: str) -> bool:
        """
        Logs into grid, creating a local credential set.

        Parameters
        ----------
        username: str
            Grid username
        key: str
            Grid API key

        Returns
        -------
        sucess: bool
            Truthy if login is successful.
        """
        #  We'll setup a new credentials header for this request
        #  and also instantiate the client.
        self.__set_authentication_headers(username=username, key=key)
        self._init_client()

        #  Let's create a directory first, using the parent
        #  path to that directory.
        P = Path.home().joinpath(self.grid_credentials_path)
        Path(P.parents[0]).mkdir(parents=True, exist_ok=True)

        query = gql("""
            query Login ($cliVersion: String!) {
                cliLogin (cliVersion: $cliVersion) {
                    userId
                    success
                    message
                }
            }
        """)

        #  Get user ID and store in credentials file.
        success = False
        params = {'cliVersion': __version__}
        try:
            result = self.client.execute(query, variable_values=params)
            credentials = {
                'UserID': result['cliLogin']['userId'],
                'APIKey': key
            }

            #  Writes JSON credentials file.
            with P.open('w') as file:
                json.dump(credentials, file, ensure_ascii=False, indent=4)

            success = True

        except HTTPError as e:
            raise click.ClickException('Failed to login')

        except Exception as e:  # skipcq: PYL-W0703
            if env.DEBUG:
                click.echo(e)
            raise click.ClickException('API error')

        return success

    def train(self, config: Dict, kind: WorkflowType, run_name: str,
              run_description: str, entrypoint: str,
              script_args: List[str]) -> None:
        """
        Submits a Run to backend from a local script.

        Parameters
        ----------
        config: str
            YAML config as a string.
        kind: WorkflowType
            Run kind; either SCRIPT or BLUEPRINT. BLUEPRINT not
            supported at the moment,
        run_name: str
            Run name
        run_description: str
            Run description.
        entrypoint: str
            Entrypoint script.
        script_args: List[str]
            Script arguments passed from command line.
        """
        # Check user Github token for user.
        self._check_user_github_token()

        # Check if the active directory is a github.com repository.
        self._check_github_repository()

        # Check if repository contains uncommitted files.
        self._check_if_uncommited_files()

        # Check if remote is in sync with local.
        self._check_if_remote_head_is_different()

        # Checking dependencies and suggesting when needed
        # TODO: do this in slurm_train as well
        change_in_deps = self._check_dependency_listing()
        if change_in_deps:
            # Inject dependencies and exit!
            self._serialize_dependencies(config)

        if kind == WorkflowType.BLUEPRINT:
            raise TrainError(
                'Blueprint workflows are currently not supported.')

        #  Base64-encode the config object either passed
        #  or constructed.
        config_encoded = base64.b64encode(
            yaml.dump(config).encode('ascii')).decode('ascii')
        env.logger.debug(config_encoded)

        #  Get commit SHA
        commit_sha = execute_git_command(['rev-parse', 'HEAD'])
        env.logger.debug(commit_sha)

        #  Get repo name
        github_repository = execute_git_command(
            ["config", "--get", "remote.origin.url"])
        env.logger.debug(github_repository)

        #  Clean up the repo name
        github_repository = github_repository.replace('git@github.com:',
                                                      'github.com/')
        github_repository = github_repository.replace('.git', '')

        #  Build GraphQL query
        mutation = gql("""
        mutation (
            $configString: String!
            $name: String!
            $description: String
            $commitSha: String
            $githubRepository: ID!
            $commandLineArgs: [String]!
            ) {
            trainScript (
                properties: {
                        githubRepository: $githubRepository
                        name: $name
                        description: $description
                        configString: $configString
                        commitSha: $commitSha
                        commandLineArgs: $commandLineArgs
                    }
            ) {
            success
            message
            name
            runId
            }
        }
        """)

        #  Add the root path to the entrypoint script.
        _entrypoint = Grid._add_git_root_path(entrypoint)

        #  Prepend the file name to the list of args and
        #  builds the query payload.
        script_args.insert(0, _entrypoint)
        params = {
            'configString': config_encoded,
            'name': run_name,
            'description': run_description,
            'commitSha': commit_sha,
            'githubRepository': github_repository,
            'commandLineArgs': script_args
        }

        #  Send request to Grid.
        try:
            result = self.client.execute(mutation, variable_values=params)
            if env.DEBUG:
                click.echo('Train response')
                click.echo(result)

        #  Raise any other errors that the backend may raise.
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            raise click.ClickException(message)

    # skipcq: PYL-W0102
    def status(self,
               kind: Optional[ObservableType] = None,
               identifiers: List[str] = None,
               follow: bool = False,
               export: str = None) -> None:
        """
        The status of an observable object in Grid. That can be a Cluster,
        a Run, or an Experiment.

        Parameters
        ----------
        kind: Optional[ObservableType], default None
            Kind of object that we should get the status from
        identifiers: List[str], default []
            Observable identifiers
        follow: bool, default False
            If we should generate a live table with results.
        export: Optional[str], default None
            What type of file results should be exported to, if any.
        """
        #  We'll instantiate a websocket client when users
        #  want to follow an observable.
        if follow:
            self._init_client(websocket=True)

        kind = kind or ObservableType.RUN

        if kind == ObservableType.EXPERIMENT:
            observable = self.available_observables[kind](
                client=self.client, identifier=identifiers[0])

        elif kind == ObservableType.RUN:
            if not identifiers:
                observable = self.available_observables[ObservableType.RUN](
                    client=self.client)
            else:
                #  For now, we only check the first observable.
                #  We should also check for others in the future.
                observable = self.available_observables[kind](
                    client=self.client, identifier=identifiers[0])

        elif kind == ObservableType.INTERACTIVE:
            # Create observable.
            observable = self.available_observables[kind](client=self.client)

        elif kind == ObservableType.CLUSTER:
            raise click.BadArgumentUsage(
                "It isn't yet possible to observe clusters.")

        else:
            raise click.BadArgumentUsage('No observable instance created.')

        if follow:
            result = observable.follow()
        else:
            result = observable.get()

        #  Save status results to a file, if the user has specified.
        if export:
            try:

                #  No need to continue if there are not results.
                if not result:
                    click.echo('\nNo run data to write to CSV file.\n')
                    return result

                #  The user may have requested a table of
                #  Runs or Experiments, use the key that is returned
                #  by the API.
                results_key = list(result.keys())[0]

                #  Initialize variables.
                path = None
                now = datetime.now()
                date_string = f'{now:%Y-%m-%d_%H:%M}'

                if export == 'csv':
                    path = f'grid-status-{date_string}.csv'
                    with open(path, 'w') as csv_file:

                        #  We'll exclude any List or Dict from being
                        #  exported in the CSV. We do this to avoid
                        #  generating a CSV that contains JSON data.
                        #  There aren't too many negative sides to this
                        #  because the nested data isn't as relevant.
                        sample = result[results_key][0]
                        _sample = sample.copy()
                        for k, v in _sample.items():
                            if isinstance(v, (list, dict)):
                                del sample[k]  # skipcq: PTC-W0043

                        columns = sample.keys()
                        writer = csv.DictWriter(csv_file, fieldnames=columns)
                        writer.writeheader()
                        for data in result[results_key]:
                            writer.writerow({
                                k: v
                                for k, v in data.items() if k in columns
                            })

                elif export == 'json':
                    path = f'grid_status-{date_string}.json'
                    with open(path, 'w') as json_file:
                        json_file.write(json.dumps(result[results_key]))

                if path:
                    click.echo(f'\nExported status to file: {path}\n')

            #  Catch possible errors when trying to create file
            #  in file system.
            except (IOError, TypeError) as e:
                if env.DEBUG:
                    click.echo(e)

                raise click.FileError('Failed to save grid status to file\n')

        return result

    # skipcq: PYL-W0102
    def history(self,
                identifiers: List[str] = [],
                kind: Optional[ObservableType] = ObservableType.RUN) -> None:
        """
        Fetches the history of an observable object in Grid. That can be a
        Cluster, a Run, or an Experiment.

        Parameters
        ----------
        kind: Optional[ObservableType], default ObservableType.RUN
            The kind of object to fetch history from
        identifiers: List[str], default []
            Object identifier, e.g. Experiment ID
        """
        if not kind:
            observable = self.available_observables[ObservableType.RUN]()

        elif kind == ObservableType.EXPERIMENT:
            observable = self.available_observables[kind](
                client=self.client, identifier=identifiers[0])

        elif kind == ObservableType.RUN:
            if not identifiers:
                observable = self.available_observables[ObservableType.RUN](
                    client=self.client)
            else:
                observable = self.available_observables[kind](
                    client=self.client, identifier=identifiers[0])

        elif kind == ObservableType.CLUSTER:
            raise click.BadArgumentUsage(
                "It isn't yet possible to observe clusters.")

        else:
            raise click.BadArgumentUsage(
                f"history not supported for kind {kind}")

        return observable.get_history()

    def _cancel_experiments(
            self,
            experiments: List[Dict[str, str]],
            spinner: Optional[yaspin.core.Yaspin] = None) -> bool:
        """
        Cancels a list of experiments.

        Parameters
        ----------
        experiments: List[Dict[str, str]]
            List of experiment objects to cancel.
        spinner:Optional[yaspin.core.Yaspin]
            yaspin spinner instance.

        Returns
        -------
        success: bool
            Truthy if operation is successful
        """
        if not spinner:
            spinner = yaspin.yaspin(
                text=f'Cancelling {len(experiments)} experiments',
                color="yellow")

        # Check that experiments are in a non cancelled status.
        success = False
        non_cancelled_statuses = ('failed', 'succeeded', 'cancelled')
        for experiment in experiments:

            result = None
            experiment_status = experiment['status']
            experiment_id = experiment['experimentId']

            if experiment_status not in non_cancelled_statuses:

                # Create a spinner for each experiment to be cancelled.
                spinner = yaspin.yaspin(text=f'Cancelling {experiment_id}',
                                        color="yellow")
                spinner.start()

                params = {'experimentId': experiment_id}
                try:
                    mutation = gql("""
                    mutation (
                        $experimentId: ID!
                    ) {
                        cancelExperiment(experimentId: $experimentId) {
                            success
                            message
                        }
                    }
                    """)
                    result = self.client.execute(mutation,
                                                 variable_values=params)

                    # Check if experiment has been cancelled successfully.
                    success = result['cancelExperiment']['success']
                    if result and success:
                        spinner.ok("✔")

                    else:
                        spinner.fail("✘")
                        spinner.stop()
                        raise click.ClickException(
                            f'Failed to cancel experiment {experiment_id}.'
                            f"{result['cancelExperiment']['message']}")

                    # Wait for T time between requests to avoid
                    # DDoSing backend.
                    time.sleep(self.request_cooldown_duration)

                except Exception as e:  # skipcq: PYL-W0703
                    spinner.fail("✘")
                    spinner.stop()
                    raise click.ClickException(
                        f"Failed to cancel experiment {experiment['experimentId']}. {e}"
                    )

                # Close spinner on every iteration.
                finally:
                    spinner.stop()

        return success

    def cancel(self,
               run_name: Optional[str] = None,
               experiment_id: Optional[str] = None) -> bool:
        """
        Cancels a run or an experiment.

        Parameters
        ----------
        run_name: Optional[str]
            Run name
        experiment_id: Optional[str]
            Experiment ID

        Returns
        -------
        success: bool
            Truthy if operation is successful.
        """
        # Create spinner for fetching experiment list.
        spinner = yaspin.yaspin(text="Loading ...", color="yellow")
        spinner.start()

        # If an experiment ID was passed, we're just
        # cancelling that experiment.
        success = False
        if experiment_id:
            experiment_data = self.experiment_details(
                experiment_id=experiment_id)

            if not experiment_data['getExperimentDetails']['status']:
                spinner.fail("✘")
                spinner.stop()
                raise click.ClickException(
                    f'Experiment {experiment_id} does not exist')

            experiments = [{
                'experimentId': experiment_id,
                'status': experiment_data
            }]

        # If a run ID was passed, we're cancelling every experiment in the run
        else:
            query = gql("""
            query (
                $runName: ID
            ) {
                getExperiments (runName: $runName) {
                    experimentId
                    status
                }
            }
            """)
            params = {'runName': run_name}

            try:
                result = self.client.execute(query, variable_values=params)
                if not result['getExperiments']:
                    raise click.ClickException(
                        f'Run {run_name} does not exist.')

            except Exception as e:  # skipcq: PYL-W0703
                literal_exception = ast.literal_eval(str(e))

                message = None
                if isinstance(literal_exception, dict):
                    message = literal_exception.get('message')

                spinner.fail("✘")
                raise click.ClickException(
                    f'Error finding run {run_name}. {message}')

            finally:
                spinner.stop()

            experiments = result['getExperiments']

        # Cancel all experiments.
        success = self._cancel_experiments(experiments=experiments,
                                           spinner=spinner)

        # Finish spinner for fetching experiments task.
        spinner.ok("✔")
        spinner.stop()

        # Add additional message if the user has cancelled a Run.
        if run_name:
            styled_run_name = click.style(run_name, fg='blue')
            click.echo(f'All experiments in Run {styled_run_name} '
                       'were cancelled successfully.')

        return success

    def create_interactive_node(self, config: str, name: str,
                                description: str) -> bool:
        """
        Creates an interactive node via Grid.

        Parameters
        ----------
        config: str
            String representation of YAML file
        name: str
            Name of interactive node to use
        description: str
            Description of interactive node

        Returns
        -------
        success: bool
            Truthy if operation is successful.
        """
        spinner = yaspin.yaspin(text="Creating Interactive node ...",
                                color="yellow")
        spinner.start()

        #  Base64-encode the config object either passed
        #  or constructed.
        config_encoded = base64.b64encode(
            config.encode('ascii')).decode('ascii')

        #  Cancel the entire Run otherwise.
        mutation = gql("""
        mutation (
            $name: ID!
            $description: String
            $configString: String!
        ) {
            createInteractiveNode(properties: {
                                    name: $name, description: $description,
                                    configString: $configString
                                  }) {
                success
                message
            }
        }
        """)

        params = {
            'name': name,
            'description': description,
            'configString': config_encoded,
        }
        try:
            result = self.client.execute(mutation, variable_values=params)
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            spinner.fail("✘")
            raise click.ClickException(message)

        success = result['createInteractiveNode']['success']
        if success:
            spinner.ok("✔")
            click.echo(f'Interactive node {name} is spinning up.')

        elif not success:
            spinner.fail("✘")
            click.echo(f"→ {result['createInteractiveNode']['message']}")

            raise click.ClickException(
                f"Failed to create interactive node '{name}'")

        return success

    def pause_interactive_node(self, interactive_node_name: str) -> None:
        """
        Pauses an interactive node.

        Parameters
        ----------
        interactive_node_name: str
            Interactive node ID
        """
        spinner = yaspin.yaspin(text="Pausing Interactive node ...",
                                color="yellow")
        spinner.start()

        mutation = gql("""
        mutation (
            $interactiveNodeName: ID!
        ) {
            pauseInteractiveNode(interactiveNodeName: $interactiveNodeName) {
                success
                message
            }
        }
        """)

        params = {'interactiveNodeName': interactive_node_name}

        success = False
        try:
            result = self.client.execute(mutation, variable_values=params)
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            spinner.fail("✘")
            raise click.ClickException(message)

        success = result['pauseInteractiveNode']['success']
        if success:
            spinner.ok("✔")
            click.echo(f'Interactive node {interactive_node_name} has ' +
                       'been paused successfully.')

        elif not success:
            spinner.fail("✘")
            if env.DEBUG:
                click.echo(f"→ {result['pauseInteractiveNode']['message']}")

            raise click.ClickException(
                f"Failed to pause interactive node '{interactive_node_name}'")

        return success

    def resume_interactive_node(self, interactive_node_name: str) -> None:
        """
        Resumes an interactive node.

        Parameters
        ----------
        interactive_node_name: str
            Interactive node ID
        """
        spinner = yaspin.yaspin(text="Resuming Interactive node ...",
                                color="yellow")
        spinner.start()

        mutation = gql("""
        mutation (
            $interactiveNodeName: ID!
        ) {
            resumeInteractiveNode(interactiveNodeName: $interactiveNodeName) {
                success
                message
            }
        }
        """)

        params = {'interactiveNodeName': interactive_node_name}

        success = False
        try:
            result = self.client.execute(mutation, variable_values=params)
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            spinner.fail("✘")
            raise click.ClickException(message)

        success = result['resumeInteractiveNode']['success']
        if success:
            spinner.ok("✔")
            click.echo(f'Interactive node {interactive_node_name} has ' +
                       'been resumed successfully.')

        elif not success:
            spinner.fail("✘")
            if env.DEBUG:
                click.echo(f"→ {result['resumeInteractiveNode']['message']}")

            raise click.ClickException(
                f"Failed to resume interactive node '{interactive_node_name}'")

        return success

    def delete_interactive_node(self, interactive_node_name: str) -> None:
        """
        Deletes an interactive node from cluster.

        Parameters
        ----------
        interactive_node_name: str
            Interactive node ID
        """
        spinner = yaspin.yaspin(text="Deleting Interactive node ...",
                                color="yellow")
        spinner.start()

        mutation = gql("""
        mutation (
            $interactiveNodeName: ID!
        ) {
            deleteInteractiveNode(interactiveNodeName: $interactiveNodeName) {
                success
                message
            }
        }
        """)

        params = {'interactiveNodeName': interactive_node_name}

        success = False
        try:
            result = self.client.execute(mutation, variable_values=params)
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            spinner.fail("✘")
            raise click.ClickException(message)

        success = result['deleteInteractiveNode']['success']
        if success:
            spinner.ok("✔")
            click.echo(f'Interactive node {interactive_node_name} has ' +
                       'been deleted successfully.')

        elif not success:
            spinner.fail("✘")
            if env.DEBUG:
                click.echo(f"→ {result['deleteInteractiveNode']['message']}")

            raise click.ClickException(
                f"Failed to delete interactive node '{interactive_node_name}'")

        return success

    def experiment_details(self, experiment_id: str) -> Dict[str, Any]:
        """
        Get experiment details.

        Parameters
        ----------
        experiment_id: str
            Experiment ID

        Returrns
        --------
        details: Dict[str, Any]
            Experiment details
        """
        # If job is queued, notify the user that logs aren't available yet
        query = gql("""
        query (
            $experimentId: ID!

        ) {
            getExperimentDetails(experimentId: $experimentId) {
                experimentId
                status
            }
        }
        """)
        params = {'experimentId': experiment_id}
        result = self.client.execute(query, variable_values=params)

        return result

    # TODO: refactor this to break into smaller methods.
    def experiment_logs(self,
                        experiment_id: str,
                        n_lines: int = 50,
                        page: Optional[int] = None,
                        max_lines: Optional[int] = None,
                        use_pager: bool = False) -> Dict[str, str]:
        """
        Gets experiment logs from a single experiment.

        Parameters
        ----------
        n_lines: int, default 200
            Max number of lines to return
        experiment_id: str
            Experiment ID for a single experiment
        page: Optional[int], default None
            Which page of logs to fetch.
        max_lines: Optional[int], default None
            Maximum number of lines to print in terminal.
        use_pager: bool, default False
            If the log results should be a scrollable pager.
        """
        #  Starts spinner.
        spinner = yaspin.yaspin(text="Fetching logs ...", color="yellow")
        spinner.start()

        # If the experiment is in a finished state, then
        # get logs from archive.
        finished_states = ('failed', 'succeeded', 'cancelled')
        experiment_details = self.experiment_details(
            experiment_id=experiment_id)

        # Check if Experiment is queued.
        state = experiment_details['getExperimentDetails']['status']

        if state is None or state == 'deleted':
            spinner.fail("✘")
            spinner.stop()
            raise click.ClickException(
                f"""Could not find experiment {experiment_id}""")

        if state == 'queued':
            spinner.ok("✔")
            styled_queued = click.style('queued', fg='yellow')
            click.echo(f"""
    Your Experiment is {styled_queued}. Logs will be available
    when your Experiment starts.
            """)
            spinner.stop()
            return

        # Check if the user has requested logs from the
        # archive explicitly or if the experiment is in a
        # finished state.
        is_archive_request = page is not None
        is_finished_state = state in finished_states
        if is_archive_request or is_finished_state:

            query = gql("""
            query GetLogs ($experimentId: ID!, $page: Int) {
                getArchiveExperimentLogs(experimentId: $experimentId, page: $page) {
                    lines {
                        message
                        timestamp
                    }
                    currentPage
                    totalPages
                }
            }
            """)
            params = {'experimentId': experiment_id, 'page': page}
            try:
                result = self.client.execute(query, variable_values=params)

            #  Raise any other errors that the backend may raise.
            except Exception as e:  # skipcq: PYL-W0703
                spinner.fail("✘")
                if 'Server error:' in str(e):
                    e = str(e).replace('Server error: ', '')[1:-1]

                message = ast.literal_eval(str(e))['message']
                raise click.ClickException(message)

            # The backend will return end empty object if no pages
            # of logs are available.
            if not result.get('getArchiveExperimentLogs') or not result[
                    'getArchiveExperimentLogs'].get('lines'):
                spinner.stop()
                raise click.ClickException(
                    f'No logs available for experiment {experiment_id}')

            # Print message to help users read all logs.
            separator = '-' * 80
            total_pages = result['getArchiveExperimentLogs']['totalPages']
            page_command_message = f"""We will be displaying logs from the archives starting
    on page 0. You can request other pages using:

        $ grid logs {experiment_id} --page 0


    Total available log pages: {total_pages}

    {separator}"""

            # Print messages indicating that other log pages are
            # available.
            styled_experiment_id = click.style(experiment_id, fg='blue')
            styled_state = click.style(state, fg='magenta')
            prompt_message = f"""

    The Experiment {styled_experiment_id} is in a finished
    state ({styled_state}). {page_command_message}

            """

            if is_archive_request:
                prompt_message = page_command_message

            spinner.ok("✔")
            spinner.stop()
            click.echo(prompt_message)

            # Get all log lines.
            lines = result['getArchiveExperimentLogs']['lines']
            total_lines = len(lines)
            if total_lines > self.acceptable_lines_to_print and not max_lines:
                styled_total_lines = click.style(str(total_lines), fg='red')
                too_many_lines_message = f"""    {click.style('NOTICE', fg='yellow')}: The log stream you requested contains {styled_total_lines} lines.
    You can limit how many lines to print by using:

        $ grid logs {experiment_id} --max_lines 50


    Would you like to proceed? """
                click.confirm(too_many_lines_message, abort=True, default=True)

            # Style the log lines.
            styled_logs = []
            for log in lines[:max_lines]:

                # If no timestamps are returned, fill the field
                # with dashes.
                if not log['timestamp']:
                    # Timestamps have 32 characters.
                    timestamp = click.style('-' * 32, fg='green')
                else:
                    timestamp = click.style(log['timestamp'], fg='green')

                styled_logs.append(f"[{timestamp}] {log['message']}")

            # Either print the logs in the terminal or use the pager
            # to scroll through the logs.
            if use_pager:
                click.echo_via_pager(styled_logs)
            else:
                for line in styled_logs:
                    click.echo(line, nl=False)

        # If the experiment isn't in a finished state, then
        # do a subscription with live logs.
        else:

            #  Let's first change the client transport to use
            #  a websocket transport instead of using the regular
            #  HTTP transport.
            self._init_client(websocket=True)

            subscription = gql("""
            subscription GetLogs ($experimentId: ID!, $nLines: Int!) {
                getLiveExperimentLogs(
                    experimentId: $experimentId, nLines: $nLines) {
                        message
                        timestamp
                }
            }
            """)

            params = {
                'experimentId':
                experiment_details['getExperimentDetails']['experimentId'],
                'nLines':
                n_lines
            }

            # Create websocket connection.
            connection_closed_message = "Connection closed. Please try again."
            try:
                stream = self.client.subscribe(subscription,
                                               variable_values=params)

                first_run = True
                for log in stream:

                    #  Closes the spinner.
                    if first_run:
                        spinner.ok("✔")
                        first_run = False

                    #  Prints each line to terminal.
                    log_entries = log['getLiveExperimentLogs']
                    for entry in log_entries:
                        # If no timestamps are returned, fill the field
                        # with dashes.
                        if not entry['timestamp']:
                            # Timestamps have 32 characters.
                            timestamp = click.style('-' * 32, fg='green')
                        else:
                            timestamp = click.style(entry['timestamp'],
                                                    fg='green')

                        click.echo(f"[{timestamp}] {entry['message']}")

            # If connection is suddenly closed, indicate that a
            # known error happened.
            except websockets.exceptions.ConnectionClosed:
                spinner.fail("✘")
                raise click.ClickException(connection_closed_message)

            except websockets.exceptions.ConnectionClosedOK:
                spinner.fail("✘")
                raise click.ClickException(
                    'Could not continue fetching log stream.')

            # Raise any other errors that the backend may raise.
            except Exception as e:  # skipcq: PYL-W0703
                raise click.ClickException(connection_closed_message)

        #  Makes sure to close the spinner in all situations.
        #  If we don't do this, the tracker character in the terminal
        #  disappears.
        spinner.stop()

    def experiment_metrics(self,
                           experiment_id: str,
                           metric: str,
                           n_lines: int = 50) -> Dict[str, str]:
        """
        Gets experiment scalar logs from a single experiment.

        Parameters
        ----------
        experiment_id: str
            Experiment ID for a single experiment
        metric: str
            Metric names
        n_lines: int, default 50
            Max number of lines to return
        """
        #  Starts spinner.
        spinner = yaspin.yaspin(text="Fetching metrics ...", color="yellow")
        spinner.start()

        experiment_details = self.experiment_details(
            experiment_id=experiment_id)

        # Check if Experiment is queued.
        state = experiment_details['getExperimentDetails']['status']
        if state == 'queued':
            spinner.ok("✔")
            styled_queued = click.style('queued', fg='yellow')
            click.echo(f"""
    Your Experiment is {styled_queued}. Metrics will be available
    when your Experiment starts.
            """)
            spinner.stop()
            return

        query = gql("""
        query GetScalarLogs ($experimentId: ID!, $metric: String!, $startIndex: Int, $format: String) {
            getExperimentScalarLogs(experimentId: $experimentId,
                                    metric: $metric,
                                    startIndex: $startIndex,
                                    format: $format) {
                values
                steps
                nextIndex
            }
        }
        """)

        params = {
            'experimentId': experiment_id,
            'metric': metric,
            'startIndex': 0,
            'format': 'json'
        }

        try:
            result = self.client.execute(query, variable_values=params)
        #  Raise any other errors that the backend may raise.
        except Exception as e:  # skipcq: PYL-W0703
            spinner.fail("✘")
            if 'Server error:' in str(e):
                e = str(e).replace('Server error: ', '')[1:-1]

            message = ast.literal_eval(str(e))['message']
            raise click.ClickException(message)

        metric = result['getExperimentScalarLogs']
        # The backend will return an empty object if no metrics
        # are available.
        if not metric or not metric['steps']:
            spinner.stop()
            raise click.ClickException(
                f'No metrics available for experiment {experiment_id}')

        spinner.ok("✔")
        spinner.stop()

        # TODO: Change when encoding and compression is fixed in
        # backend API
        # def decode(x: bytes, dtype: str):
        #     buf = zlib.decompress(base64.b64decode(x.encode()))
        #     out = array(dtype)
        #     out.frombytes(buf)
        #     return out

        try:
            # steps = decode(metric['steps'], 'q')
            # values = decode(metric['values'], 'd')

            steps = json.loads(metric['steps'])
            values = json.loads(metric['values'])
        #  Raise any other errors that the backend may raise.
        except Exception as e:
            spinner.fail("✘")
            if 'Server error:' in str(e):
                e = str(e).replace('Server error: ', '')[1:-1]

            message = ast.literal_eval(str(e))['message']
            raise click.ClickException(message)

        if n_lines > 0:
            steps = steps[-n_lines:]
            values = values[-n_lines:]

        # Get all metrics in lines
        lines = [
            f"{step:>16}  {value:.10f}\n"
            for step, value in zip(steps, values)
        ]

        # Print the metrics in the terminal
        for line in lines:
            click.echo(line, nl=False)

        #  Makes sure to close the spinner in all situations.
        #  If we don't do this, the tracker character in the terminal
        #  disappears.
        spinner.stop()

    def download_experiment_artifacts(self, experiment_id: str,
                                      download_dir: str) -> None:
        """
        Downloads artifacts for a given experiment.
        Parameters
        ----------
        experiment_id: str
            Experiment ID for artifact.
        download_dir: str
            Download path
        """
        #  Starts spinner.
        spinner = yaspin.yaspin(text="Downloading artifacts ...",
                                color="yellow")
        spinner.start()

        mutation = gql("""
        query (
            $experimentId: ID!
        ) {
            getArtifacts(experimentId: $experimentId) {
                signedUrl
                downloadToPath
                downloadToFilename
            }
        }
        """)

        # Make request and catch any possible errors with the actual query.
        params = {'experimentId': experiment_id}
        try:
            result = self.client.execute(mutation, variable_values=params)
            spinner.ok("✔")
            click.echo(f'Starting download for: {experiment_id}')
        except Exception as e:  # skipcq: PYL-W0703
            spinner.fail("✘")
            message = ast.literal_eval(str(e))['message']
            raise click.ClickException(str(message))

        # Create host directory.
        Downloader.create_dir_tree(download_dir)

        # Create downloadable objects.
        files_to_download = []
        if result['getArtifacts']:
            for artifact in result['getArtifacts']:
                files_to_download.append(
                    DownloadableObject(
                        url=artifact['signedUrl'],
                        download_path=artifact['downloadToPath'],
                        filename=artifact['downloadToFilename']))

            # Start download if there are any files to download.
            if files_to_download:
                D = Downloader(downloadable_objects=files_to_download,
                               base_dir=download_dir)
                D.download()

            # Display message to users indicating that experiment
            # has no artifacts.
            else:
                click.echo(f'Experiment {experiment_id} has no artifacts.')

        #  Close spinner.
        spinner.stop()

    def delete(self,
               experiment_id: Optional[str] = None,
               run_id: Optional[str] = None) -> bool:
        """
        Deletes an experiment or a run

        Parameters
        ----------
        experiment_id : Optional[str]
            experiment ID of experiment to be deleted
        run_id : Optional[str]
            run ID of run to be deleted

        Returns
        -------
        success: bool
            Truthy if operation is successful
        """
        # Create spinner
        spinner = yaspin.yaspin(
            text=f'Deleting {experiment_id if experiment_id else run_id}',
            color="yellow")
        spinner.start()
        success = False

        if experiment_id:
            params = {'experimentId': experiment_id}
            mutation = gql("""
            mutation (
                $experimentId: ID!
            ) {
                deleteExperiment(experimentId: $experimentId) {
                    success
                    message
                }
            }
            """)
            try:
                result = self.client.execute(mutation, variable_values=params)
                success = result['deleteExperiment']['success']
                if result and success:
                    spinner.ok("✔")
                    click.echo(
                        f'Experiment {experiment_id} has been deleted successfully'
                    )
                else:
                    spinner.fail("✘")
                    raise click.ClickException(
                        f"Failed to delete experiment {experiment_id}. "
                        f"{result['deleteExperiment']['message']}")
            except Exception as e:  # skipcq: PYL-W0703
                spinner.fail("✘")
                raise click.ClickException(
                    f'Failed to delete experiment {experiment_id}. {e}')

        else:
            params = {'name': run_id}
            mutation = gql("""
            mutation (
                $name: ID!
            ) {
                deleteRun(name: $name) {
                    success
                    message
                }
            }
            """)
            try:
                result = self.client.execute(mutation, variable_values=params)
                success = result['deleteRun']['success']
                if result and success:
                    spinner.ok("✔")
                    click.echo(f'Run {run_id} has been deleted successfully')
                else:
                    spinner.fail("✘")
                    raise click.ClickException(
                        f"Failed to delete run {run_id}."
                        f"{result['deleteRun']['message']}")
            except Exception as e:  # skipcq: PYL-W0703
                spinner.fail("✘")
                raise click.ClickException(
                    f'Failed to delete run {run_id}. {e}')

        spinner.stop()
        return success

    def delete_datastore(self, name: str, version: int, credential_id: str):
        """
        Delete datastore for user

        Parameters
        ----------
        name: str
            Datastore name
        version: int
            Datastore version
        credential_id: str
            Credential Id
        """
        mutation = gql("""
            mutation (
                $name: String!
                $version: Int!
                $credentialId: String!
                ) {
                deleteDatastore (
                    properties: {
                            name: $name,
                            version: $version,
                            credentialId: $credentialId
                        }
                ) {
                success
                message
                }
            }
            """)

        params = {
            'name': name,
            'version': version,
            'credentialId': credential_id
        }

        spinner = yaspin.yaspin(
            text=f'Deleting datastore {name} with version {version}',
            color="yellow")
        spinner.start()

        try:
            result = self.client.execute(mutation, variable_values=params)
            success = result['deleteDatastore']['success']
            message = result['deleteDatastore']['message']

            if success:
                spinner.text = "Finished deleting datastore."
                spinner.ok("✔")
            else:
                spinner.text = f'Failed to delete datastore {name} with version ' \
                               f'{version}: {message}'
                spinner.fail("✘")

        except Exception as e:  #skipcq: PYL-W0703
            spinner.fail("✘")
            raise click.ClickException(
                f'Failed to delete datastore {name} with version {version}. {e}'
            )

        spinner.stop()
        return success

    def resume_datastore_session(self, session_name: str):
        """
        Resume uploading a datastore session

        Parameters
        ----------
        session_name: str
            Name of session
        """
        sessions = DatastoreUploadSession.recover_sessions()
        session = None
        for s in sessions:
            if s.session_name == session_name:
                session = s
                break

        if not session:
            self.__print_datastore_sessions(sessions)
            raise click.ClickException(
                f"No session named {session_name} found, please specify an " +
                "available session")

        session.configure(client=self.client)
        try:
            session.upload()
        except Exception as e:  #skipcq: PYL-W0703
            raise click.ClickException(
                f'Failed to upload datastore {session.name}. {e}')

    @staticmethod
    def __print_datastore_sessions(sessions: List[DatastoreUploadSession]):
        """
        Print the table of datastore sessions
        """
        click.echo(
            "Available datastore incomplete uploads that can be resumed: ")
        table_cols = ['Name', 'Version', "Command to resume"]
        table = BaseObservable.create_table(columns=table_cols)
        for session in sessions:
            command = f"grid datastores resume {session.session_name}"
            table.add_row(session.name, str(session.version), command)

        console = Console()
        console.print(table)

    def list_resumable_datastore_sessions(self):
        """
        List datastores that can be resumed upload
        """
        sessions = DatastoreUploadSession.recover_sessions()
        if len(sessions) == 0:
            return

        click.echo("")
        click.echo("")
        self.__print_datastore_sessions(sessions)

    def validate_datastore_version(self, grid_config: dict):
        """
        Find maximum version of a datastore (ready to attach) and set config accordingly.
        No-op if config does not contain datastore name or already contains a set datastore version.
        """
        try:
            datastore_name = grid_config["compute"]["train"]["datastore_name"]
            datastore_version = grid_config["compute"]["train"][
                "datastore_version"]
        except:
            return

        if datastore_version or not datastore_name:
            return

        query = gql("""
            query GetDatastores {
                getDatastores {
                    name
                    version
                    snapshotStatus
                }
            }
        """)

        try:
            result = self.client.execute(query)
        except TransportQueryError as e:
            raise click.ClickException(f"Unable to list datastores: {e}")

        max_version = -1
        for row in result["getDatastores"]:
            if datastore_name == row["name"] and row[
                    'snapshotStatus'] == "succeeded":
                max_version = max(max_version, int(row["version"]))

        if max_version == -1:
            raise click.ClickException(
                f'No --grid_datastore_version passed for datastore: {datastore_name}, but unable to find a ready-to-use version.'
            )

        click.echo(
            f'No --grid_datastore_version passed for datastore: {datastore_name}. Using version: {max_version}'
        )
        grid_config["compute"]["train"]["datastore_version"] = str(max_version)

    def list_datastores(self):
        """
        List datastores for user
        """
        query = gql("""
            query GetDatastores {
                getDatastores {
                    id
                    credentialId
                    name
                    version
                    size
                    createdAt
                    snapshotStatus
                }
            }
        """)
        try:
            result = self.client.execute(query)
        except TransportQueryError as e:
            raise click.ClickException(f"Unable to list datastores: {e}")

        table_cols = [
            "Credential Id", "Name", "Version", "Size", "Created", "Status"
        ]
        table = BaseObservable.create_table(columns=table_cols)
        for row in result["getDatastores"]:
            created_at = date_string_parse(row["createdAt"])
            created_at = f'{created_at:%Y-%m-%d %H:%M}'
            size = row["size"]
            if size or size == 0:
                size = humanize.naturalsize(size * (1024**2))

            if row['snapshotStatus'] in {
                    'pending', 'preparing', 'queued', 'running'
            }:
                status = 'Optimizing'
            else:
                status = row['snapshotStatus'].title(
                ) if row['snapshotStatus'] else "Unknown"

            table.add_row(row["credentialId"], row["name"], row["version"],
                          size, created_at, status)

        console = Console()
        console.print(table)

    def upload_datastore(self, source: str, name: str, credential_id: str,
                         compression: bool):
        """
        Uploads datastore to storage

        Parameters
        ----------
        source: str
           Source to create datastore from, either a local directory or a
           http url.
        name: str
           Name of datastore
        credential_id: str
            Grid credential id
        compression: str
            Enable compression, which is disabled by default

        Returns
        -------
        success: bool
            Truthy when a datastore has been uploaded
            correctly.
        """
        try:
            session = create_datastore_session(name=name,
                                               source=source,
                                               credential_id=credential_id,
                                               compression=compression,
                                               client=self.client)
            session.upload()
        except Exception as e:  #skipcq: PYL-W0703
            raise click.ClickException(
                f'Failed to upload datastore {name}. {e}')

        return True

    def get_slurm_auth_token(self, alias: str = None):
        """
        Gets Slurm auth token and alias to use for registering
        a daemon in a users SLURM cluster.

        Parameters
        ----------
        alias: str
            An optional user defined alias
            to give their daemon.
        """
        spinner = yaspin.yaspin(text='Generating token for grid-daemon use.',
                                color="yellow")
        spinner.start()

        try:
            query = gql("""
            query ($alias: String) {
                getSlurmAuthToken (
                    alias: $alias
                ) {
                success
                message
                token
                alias
                }
            }
            """)

            params = {'alias': alias}

            result = self.client.execute(query, variable_values=params)
            success = result['getSlurmAuthToken']['success']

            if success:
                token = result['getSlurmAuthToken']['token']
                alias = result['getSlurmAuthToken']['alias']

                spinner.text = "Finished generating token."
                spinner.ok("✔")

                click.echo(f"Token: {token}")
                click.echo(f"Alias: {alias}")
            else:
                spinner.fail("✘")

                message = result['getSlurmAuthToken']['message']
                raise click.ClickException(
                    f'Failed to create auth token. {message}')

        except Exception as e:  #skipcq: PYL-W0703
            spinner.fail("✘")
            raise click.ClickException(f'Failed to create auth token. {e}')

        spinner.stop()
        return success

    def train_on_slurm(self, config_str: str, run_name: str,
                       script_args: Optional[Any]):
        """
        Gets Slurm auth token and alias to use for registering
        a daemon in a users SLURM cluster.

        Parameters
        ----------
        alias: str
            An optional user defined alias
            to give their daemon.
        """
        # Check user Github token for user.
        self._check_user_github_token()

        # Check if the active directory is a github.com repository.
        self._check_github_repository()

        # Check if repository contains uncommited files.
        self._check_if_uncommited_files()

        # Check if remote is in sync with local.
        self._check_if_remote_head_is_different()

        #  Base64-encode the config object either passed
        #  or constructed.
        config_encoded = base64.b64encode(
            config_str.encode('ascii')).decode('ascii')
        env.logger.debug(config_encoded)

        #  Get commit SHA
        commit_sha = execute_git_command(['rev-parse', 'HEAD'])
        env.logger.debug(commit_sha)

        #  Get repo name
        github_repository = execute_git_command(
            ["config", "--get", "remote.origin.url"])
        env.logger.debug(github_repository)

        #  Clean up the repo name
        github_repository = github_repository.split('/')[1]
        github_repository = github_repository.replace('.git', '')

        try:
            #  Build GraphQL query
            mutation = gql("""
            mutation (
                $configString: String!
                $name: String!
                $description: String
                $commitSha: String
                $githubRepository: ID!
                $commandLineArgs: [String]!
                ) {
                trainSlurmScript (
                    properties: {
                            githubRepository: $githubRepository
                            name: $name
                            description: $description
                            configString: $configString
                            commitSha: $commitSha
                            commandLineArgs: $commandLineArgs
                        }
                ) {
                success
                message
                }
            }
            """)

            params = {
                "githubRepository": github_repository,
                "name": run_name,
                "configString": config_encoded,
                "commitSha": commit_sha,
                "commandLineArgs": script_args
            }

            result = self.client.execute(mutation, variable_values=params)
            success = result['trainSlurmScript']['success']

            if success:
                click.echo(result['trainSlurmScript']['message'])
            else:
                message = result['trainSlurmScript']['message']
                click.ClickException(f'Failed to train on slurm. {message}')

        except Exception as e:  #skipcq: PYL-W0703
            raise click.ClickException(f'Failed to start training. {e}')

        return result['trainSlurmScript']

    def slurm_status(self, cluster_alias: str, run_name: Optional[str] = None):
        """
        Gets statuses from Slurm.

        Parameters
        ----------
        cluster_alias: str
            An optional user defined alias
            to give their daemon.
        run_name: Optional[str]
            The name of the run
            to get a status for.
        """
        spinner = yaspin.yaspin(
            text='Getting the status from your slurm cluster.', color="yellow")
        spinner.start()
        try:
            #  Build GraphQL query
            query = gql("""
            query (
                $clusterAlias: String!
                $runName: String
                ) {
                getSlurmStatus (
                    clusterAlias: $clusterAlias
                    runName: $runName
                ) {
                success
                message,
                output
                }
            }
            """)

            params = {"clusterAlias": cluster_alias, "runName": run_name}

            result = self.client.execute(query, variable_values=params)
            success = result['getSlurmStatus']['success']

            if success:
                spinner.text = "Finished getting status from cluster."
                spinner.ok("✔")

                click.echo(result['getSlurmStatus']['message'])

                output = result['getSlurmStatus']['output']
                click.echo(output)
            else:
                spinner.fail("✘")

                message = result['getSlurmStatus']['message']
                click.echo(f'Failed to get status. {message}')

        except Exception as e:  #skipcq: PYL-W0703
            spinner.fail("✘")
            raise click.ClickException(f'Failed to send status request. {e}')

        return result['getSlurmStatus']

    def add_ssh_public_key(self, key: str, name: str):
        return self.execute_gql("""
            mutation (
                $publicKey: String!
                $name: String!
            ) {
            addSSHPublicKey(name: $name, publicKey: $publicKey) {
                message
                success
                id
              }
            }
        """,
                                publicKey=key,
                                name=name)

    def list_public_ssh_keys(self, limit: int) -> List[Dict[str, str]]:
        result = self.execute_gql(
            """
            query (
                $limit: Int!,
            ) {
                getPublicSSHKeys(limit: $limit) {
                    id
                    publicKey,
                    name
              }
            }
        """,
            limit=limit,
        )
        return result['getPublicSSHKeys']

    def delete_ssh_public_key(self, key_id: str):
        self.execute_gql("""
            mutation (
                $id: ID!
            ) {
              deleteSSHPublicKey(id: $id) {
                message
                success
              }
            }
        """,
                         id=key_id)

    def list_interactive_node_ssh_setting(self):
        return self.execute_gql("""
        query {
            getInteractiveNodes {
                name: interactiveNodeName
                ssh_url: sshUrl
            }
        }
        """)['getInteractiveNodes']

    @staticmethod
    def _gen_ssh_config(
        interactive_nodes: List[Dict[str, Any]],
        curr_config: str,
        start_marker: str = "### grid.ai managed BEGIN do not edit manually###",
        end_marker: str = "### grid.ai managed END do not edit manually###",
    ):

        content = curr_config.splitlines()
        sol = []
        managed_part = [start_marker]
        for node in interactive_nodes:
            ssh_url = urlsplit(node['ssh_url'])
            managed_part.extend([
                f"Host {node['name']}",
                f"    User {ssh_url.username}",
                f"    Hostname {ssh_url.hostname}",
                f"    Port {ssh_url.port}",
                "    StrictHostKeyChecking accept-new",
                "    CheckHostIP no",
            ])
        managed_part.append(end_marker)

        within_section = False
        added_managed_part = False
        for line in content:
            if line == start_marker:
                if added_managed_part:
                    raise ValueError("Found 2 start markers")
                if within_section:
                    raise ValueError("Found 2 start markers in row")
                within_section = True
            elif end_marker == line:
                if added_managed_part:
                    raise ValueError("Found 2+ start end")
                if not within_section:
                    raise ValueError("End marker before start marker")
                within_section = False
                sol.extend(managed_part)
                added_managed_part = True
            elif not within_section:
                sol.append(line)
        if within_section:
            raise ValueError("Found only start marker, no end one found")
        if not added_managed_part:
            sol.extend(managed_part)
        return '\n'.join(sol)

    def sync_ssh_config(self) -> List[str]:
        """
        sync local ssh config with grid's interactive nodes config

        Returns
        -------
        list of interactive nodes present
        """
        ixNodes = self.list_interactive_node_ssh_setting()
        ssh_config = Path(env.GRID_SSH_CONFIG)
        if not ssh_config.exists():
            ssh_config.write_text("")
        ssh_config.write_text(
            self._gen_ssh_config(
                interactive_nodes=ixNodes,
                curr_config=ssh_config.read_text(),
            ))
        return [x['name'] for x in ixNodes]



from typing import Optional

import click

from grid import Grid
import grid.globals as env
from grid.types import ObservableType
from grid.utilities import is_experiment


@click.command()
@click.argument('run', type=str, nargs=1, required=False)
@click.option('--follow',
              type=bool,
              help='Follows Run setup progress in terminal',
              is_flag=True)
@click.option('--details',
              type=bool,
              help='Shows Run submission details',
              is_flag=True)
@click.option('--export',
              type=click.Choice(['csv', 'json'], case_sensitive=False),
              help='Exports status output to supplied file type')
def status(run: Optional[str] = None,
           follow: Optional[bool] = False,
           details: Optional[bool] = False,
           export: Optional[str] = None) -> None:
    """Checks the status of Runs and Experiments."""
    client = Grid()

    #  Setup global flag to show build errors with
    #  Runs or Experiments. This is more effective than
    #  passing the flag around through all the invocations.
    env.SHOW_PROCESS_STATUS_DETAILS = details

    # Users can see the statuses of Experiments
    # by selecting runs.
    kind = ObservableType.RUN
    observable = []
    if run:
        observable = [run]
        kind = ObservableType.EXPERIMENT

        # Users cannot see the status of experiments directly,
        # but need to check the status of a specific run to
        # see the status of experiments.
        if is_experiment(observable[0]):
            raise click.BadArgumentUsage(
                'You can only check the status of Runs')

    client.status(kind=kind,
                  identifiers=observable,
                  follow=follow,
                  export=export)

    # If we have a Run, then don't print the global
    # interactive nodes table.
    if not run:
        client.status(kind=ObservableType.INTERACTIVE,
                      identifiers=observable,
                      follow=follow,
                      export=export)



"""
Adapted from rich's example at:

    * https://github.com/willmcgugan/rich/blob/master/examples/downloader.py
"""
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import partial
import os.path
from pathlib import Path
import signal
from threading import Event
from time import sleep
from typing import List, Optional
from urllib.request import urlopen

from rich.progress import BarColumn
from rich.progress import DownloadColumn
from rich.progress import Progress
from rich.progress import TaskID
from rich.progress import TextColumn
from rich.progress import TimeRemainingColumn
from rich.progress import TransferSpeedColumn

# Handles interruption events.
DONE_EVENT = Event()


def _handle_sigint(*args):
    DONE_EVENT.set()


# signal.signal(signal.SIGKILL, _handle_sigint)
signal.signal(signal.SIGINT, _handle_sigint)


@dataclass
class DownloadableObject:
    """Object that can be downloaded from an URL into a local path."""
    url: str
    download_path: str
    filename: str


class Downloader:
    """
    Downloads multiple files into their respective location
    using background threads.

    Attributes
    ----------
    workers: int
        Number of background threads to use.
    progress: Progress
        Instance of a Progress bar object.
    Parameters
    ----------
    urls: List[DownloadableObject]
        List of URL objects containing.
    """
    workers: int = 4

    # Creates the layout for a progress bar.
    progress = Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(bar_width=None),
        "[self.progress.percentage]{task.percentage:>3.1f}%", "•",
        DownloadColumn(), "•", TransferSpeedColumn(), "•",
        TimeRemainingColumn())

    def __init__(self, downloadable_objects: List[DownloadableObject],
                 base_dir: str):
        self.downloadable_objects = downloadable_objects
        self.base_dir = base_dir

    def _get_file(self, task_id: TaskID, url: str, path: str) -> None:
        """Copy data from a url to a local file."""
        # skipcq: BAN-B310
        response = urlopen(url)

        # This will break if the response doesn't contain content length.
        # Grid objects always have Content-length
        content_length = int(response.info()["Content-Length"])
        self.progress.update(task_id, total=content_length)
        with open(path, "wb") as dest_file:
            self.progress.start_task(task_id)
            for data in iter(partial(response.read, 32768), b""):
                dest_file.write(data)
                self.progress.update(task_id, advance=len(data))

                # Cancel downlad midway
                if DONE_EVENT.is_set():
                    return

    @staticmethod
    def create_dir_tree(dest_dir: str, base_dir: Optional[str] = None) -> None:
        """
        Creates directory structure for downloading file.

        Parameters
        ----------
        dest_dir: str
            Destination directory for where to download file.
        base_dir: str
            Base directory to place all targer directories into.
        """
        if base_dir:
            P = Path(base_dir) / Path(dest_dir)
        else:
            P = Path(dest_dir)

        P.mkdir(parents=True, exist_ok=True)

    def download(self):
        """
        Download multuple files to the given directory. This will download
        files using a ThreadPoolExecutor so that multiple files can
        be downloaded concurrently. We can set the concurrency level
        by changing the class' workers` attribute.
        """
        with self.progress:
            with ThreadPoolExecutor(max_workers=self.workers) as pool:
                for file_to_download in self.downloadable_objects:
                    Downloader.create_dir_tree(file_to_download.download_path,
                                               self.base_dir)
                    dest_path = os.path.join(self.base_dir,
                                             file_to_download.download_path,
                                             file_to_download.filename)
                    task_id = self.progress.add_task(
                        'download',
                        filename=file_to_download.filename,
                        start=False)

                    pool.submit(self._get_file, task_id, file_to_download.url,
                                dest_path)

                    if DONE_EVENT.is_set():
                        return

        return True


# NOTE: I am leaving this here to improve developer UX
# and debugability for this area. I'm inclined to refactor
# this as stated here: CORE-2644
# if __name__ == '__main__':
#     url = 'https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_272x92dp.png'
#     directory_path = 'nested_0/nested_1'
#     filename = 'test-filename.png'
#     objects = []

#     for i in range(10):
#         objects.append(
#         DownloadableObject(url=url,
#                             download_path=directory_path,
#                             filename=filename))

#     # Download file
#     D = Downloader(downloadable_objects=objects, base_dir='foo')
#     D.download()



from click.testing import CliRunner
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
from grid.client import Grid

RUNNER = CliRunner()
create_test_credentials()


def monkey_patch_download_artifacts(_self, *args, **kwargs):
    """Monkey patch the client download."""
    return


def test_artifacts_succeeds(monkeypatch):
    """grid train without arguments fails"""
    monkeypatch.setattr(Grid, '_init_client', monkey_patch_client)
    monkeypatch.setattr(Grid, 'download_experiment_artifacts',
                        monkey_patch_download_artifacts)

    result = RUNNER.invoke(cli.artifacts, ['foo-bar-exp0'])
    assert result.exit_code == 0
    assert not result.exception


def test_artifacts_raises_exception_with_non_experiments(monkeypatch):
    """grid artifacts raises exception when user passes non-experiments"""
    monkeypatch.setattr(Grid, '_init_client', monkey_patch_client)
    monkeypatch.setattr(Grid, 'download_experiment_artifacts',
                        monkey_patch_download_artifacts)

    result = RUNNER.invoke(cli.artifacts, ['foo-bar'])
    assert result.exit_code == 2
    assert result.exception


def test_artifacts_raises_exception_with_file_as_path(monkeypatch):
    """
    grid artifacts raises exception when user passes a file instead of
    a path to a directory
    """
    monkeypatch.setattr(Grid, '_init_client', monkey_patch_client)
    monkeypatch.setattr(Grid, 'download_experiment_artifacts',
                        monkey_patch_download_artifacts)

    result = RUNNER.invoke(cli.artifacts,
                           ['foo-bar', '--download_dir', 'foo.txt'])
    assert result.exit_code == 2
    assert result.exception



import json
from typing import Any, Dict, List, Optional

import click
from dateutil.parser import parse as date_string_parse
from rich.console import Console

from grid import Grid
from grid.observables import BaseObservable


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('--set_default', type=str, help='Credential set to make default')
def credentials(ctx, set_default: Optional[str] = None) -> None:
    """Manages credentials in Grid"""
    client = Grid()

    if ctx.invoked_subcommand is None:
        if set_default:
            client.credentials_update(credential_id=set_default,
                                      is_default=True)

        result = client.get_credentials()
        print_creds_table(result['getUserCredentials'])


@credentials.command()
@click.option('--provider',
              type=click.Choice(['aws'], case_sensitive=False),
              help='Credential provider.',
              required=True)
@click.option('--file',
              type=click.File('r'),
              help='JSON file to where credentials are',
              required=True)
@click.option('--alias', type=str, help='Given name for a credential set')
@click.option('--description',
              type=str,
              help='Description for a credential set')
def add(provider: Optional[str] = None,
        file: Optional[click.File] = None,
        alias: Optional[str] = None,
        description: Optional[str] = None) -> None:
    """Adds credentials in Grid"""
    client = Grid()

    #  Creates new credentials.
    creds = json.load(file)
    client.credentials_add(provider=provider,
                           credentials=creds,
                           alias=alias,
                           description=description)

    #  Print credentials in terminal.
    result = client.get_credentials()
    print_creds_table(result['getUserCredentials'])


def print_creds_table(creds: List[Dict[str, Any]]) -> Console:
    """Prints table with cloud credentials."""

    if not creds:
        message = """
            No credentials available. Add new credentials with:

                grid credentials add --file [CREDENTIALS_JSON_FILE] --provider [PROVIDER]

            """
        raise click.ClickException(message)
    table = BaseObservable.create_table(columns=[
        'Credential', 'Provider', 'Alias', 'Created At', 'Last Used At',
        'Default'
    ])

    for row in creds:
        created_at = date_string_parse(row['createdAt'])
        _created_at = f'{created_at:%Y-%m-%d %H:%M}'
        table.add_row(row['credentialId'], row['provider'], row['alias'],
                      _created_at, row['lastUsedAt'],
                      str(row['defaultCredential']))

    console = Console()
    console.print(table)

    #  We return the console for testing purposes.
    #  This isn't actually used anywhere else.
    return console



class AuthenticationError(Exception):
    """Risen when the user is not authenticated."""
    pass


class TrainError(Exception):
    """
    Risen whenever we have an exception during a training
    operation.
    """
    pass



from click.testing import CliRunner
from tests.mock_backend import resolvers
from tests.utilities import create_test_credentials
from tests.utilities import monkey_patch_client

from grid import cli
import grid.client as grid

RUNNER = CliRunner()


class TestCancel:
    @classmethod
    def setup_class(cls):
        grid.Grid._init_client = monkey_patch_client
        grid.gql = lambda x: x

        create_test_credentials()

    def test_cancel_without_arguments(self):
        """grid cancel without arguments fails"""
        result = RUNNER.invoke(cli.cancel, [])
        assert result.exit_code == 2
        assert result.exception

    def test_cancel_experiment(self):
        """grid cancel experiments with single and multi-input"""
        experiments = [f'test-experiment-exp{i}' for i in range(5)]

        result = RUNNER.invoke(cli.cancel, experiments[0])
        assert result.exit_code == 0
        assert not result.exception

        result = RUNNER.invoke(cli.cancel, experiments)
        assert result.exit_code == 0
        assert not result.exception

    def test_cancel_runs(self):
        """grid cancel Runs with single and multi-input"""
        runs = [f'test-run-{i}' for i in range(5)]
        result = RUNNER.invoke(cli.cancel, runs[0])
        assert result.exit_code == 0
        assert not result.exception

        result = RUNNER.invoke(cli.cancel, runs)
        assert result.exit_code == 0
        assert not result.exception

    def test_cancel_run_fails_with_no_experiments(self, monkeypatch):
        """grid cancel Run fails if Run has no experiments."""
        test_run = 'test-no-experiments-run'

        def mp_get_experiments(*args, **kwargs):
            return []

        monkeypatch.setattr(resolvers, 'get_experiments', mp_get_experiments)
        result = RUNNER.invoke(cli.cancel, test_run)
        assert result.exit_code == 1
        assert result.exception

    def test_cancel_run_fails_because_run_does_not_exist(self, monkeypatch):
        """grid cancel Run fails if Run does not exist."""
        test_run = 'test-no-experiments-run'

        def mp_get_experiments(*args, **kwargs):
            return {'error': 'Run does not exist'}

        monkeypatch.setattr(resolvers, 'get_experiments', mp_get_experiments)
        result = RUNNER.invoke(cli.cancel, test_run)
        assert result.exit_code == 1
        assert result.exception

    def test_only_cancels_experiments_in_not_terminal_state(self, monkeypatch):
        """grid cancel skips experiments in a terminal state."""

        not_cancelled_experiments = [
            'test-run-exp0', 'test-run-exp1', 'test-run-exp2'
        ]
        cancelled_experiment = 'test-run-exp3'

        def mp_get_experiments(*args, **kwargs):
            return [{
                'experimentId': not_cancelled_experiments[0],
                'status': 'failed'
            }, {
                'experimentId': not_cancelled_experiments[1],
                'status': 'succeeded'
            }, {
                'experimentId': not_cancelled_experiments[2],
                'status': 'cancelled'
            }, {
                'experimentId': cancelled_experiment,
                'status': 'running'
            }]

        monkeypatch.setattr(resolvers, 'get_experiments', mp_get_experiments)
        result = RUNNER.invoke(cli.cancel, 'test-run')
        assert result.exit_code == 0
        assert cancelled_experiment in result.output
        assert all(e not in result.output for e in not_cancelled_experiments)



import ast
from typing import List

import click
from gql import gql
from rich.console import Console
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TextColumn
from rich.table import Table
from yaspin import yaspin

import grid.globals as env

#  Maps backend class types to user-friendly messages.
TASK_CLASS_MAPPING = {
    'grid.core.repository_builder.RepositoryBuilder': 'Building container',
    'grid.core.cluster.Cluster': 'Creating cluster',
    'grid.core.trainer.experiment.Experiment': 'Scheduling experiment',
    'grid.core.trainer.run.RunNodePool': 'Creating node pool',
    'grid.core.trainer.interactive.InteractiveNodeTask':
    'Creating interactive node',
    'grid.core.trainer.experiment.ExperimentsWarpSpeed':
    'Scheduling experiment'
}

# Backend classes that we don't show to users. These
# are automatically triggered if a given user does not
# have global properties set.
TASK_CLASS_IGNORE = (
    'grid.core.clusters.deploy_tensorboard.ReconcileTensorboard',
    'grid.core.clusters.global_user_cluster.ReconcileCluster',
    'grid.core.user.ReconcileUser')


def style_status(format_string: str, status: str):
    """
    Styles a status message using click.stye.

    Parameters
    ----------
    status: str
        Status message to style.

    Return
    ------
    styled_status: str
        Styled string
    """
    styled_status = format_string

    if status == 'failed':
        styled_status = click.style(styled_status, fg='red')
    elif status in ('finished', 'ready'):
        styled_status = click.style(styled_status, fg='green')
    elif status in ('running', 'queued', 'pending'):
        styled_status = click.style(styled_status, fg='yellow')
    elif status in ('cancelled'):
        styled_status = click.style(styled_status, fg='white')

    return styled_status


class BaseObservable:
    def __init__(self, client, spinner_load_type=""):
        self.client = client
        self.console = Console()
        self.spinner = yaspin(text=f"Loading {spinner_load_type}...",
                              color="yellow")

    @staticmethod
    def create_table(columns: List[str]) -> Table:
        table = Table(show_header=True, header_style="bold green")

        table.add_column(columns[0], style='dim')
        for column in columns[1:]:
            table.add_column(column, justify='right')

        return table

    def _get_task_run_dependencies(self, run_name: str):
        """Gets dependency data for a given Run"""
        query = gql("""
        query (
            $runName: ID!
        ) {
            getRunTaskStatus (
                runName: $runName
            ) {
                success
                runId
                name
                status
                message
                dependencies {
                    taskId
                    status
                    taskType
                    message
                    error
                }
            }
        }
        """)
        params = {'runName': run_name}

        #  Make GraphQL query.
        result = None
        try:
            result = self.client.execute(query, variable_values=params)
            if not result['getRunTaskStatus']['success']:
                raise Exception(result['getRunTaskStatus'])
        except Exception as e:  # skipcq: PYL-W0703
            message = ast.literal_eval(str(e))['message']
            self.spinner.fail("✘")
            self.spinner.stop()

            if env.DEBUG:
                click.echo(message)

            if 'not found' in message or 'No runs available' in message:
                raise click.ClickException(
                    f'Run {run_name} not found. Did you cancel it already?')

        if result:
            dependencies = result['getRunTaskStatus']['dependencies']
            return dependencies

    def _get_task_run_status(self, run_name: str):
        #  Get dependency data.
        dependencies = self._get_task_run_dependencies(run_name=run_name)

        #  Dict to collect all errors for given tasks.
        dependency_data = {
            k: {
                'statuses': [],
                'errors': [],
                'messages': []
            }
            for k in TASK_CLASS_MAPPING if k not in TASK_CLASS_IGNORE
        }
        for task in dependencies:
            if task['taskType'] in TASK_CLASS_IGNORE:
                continue
            dependency_data[task['taskType']]['statuses'].append(
                task['status'])
            dependency_data[task['taskType']]['errors'].append(task['error'])
            dependency_data[task['taskType']]['messages'].append(
                task['message'])

        #  Inform user that she can see error logs by passing
        #  a flag.
        all_statuses = []
        for status in dependency_data.values():
            all_statuses += status['statuses']

        if any(s == 'failed' for s in all_statuses) and \
            not env.SHOW_PROCESS_STATUS_DETAILS:
            click.echo(f'''
        The Run "{run_name}" failed to start due to a setup error.
        You can see the errors by running

            grid status {run_name} --details

            ''')

        #  If there's an error with and pre-run steps, then
        #  print that error to the terminal.
        if env.SHOW_PROCESS_STATUS_DETAILS:
            for key, value in dependency_data.items():
                for error, message, status in zip(value['errors'],
                                                  value['messages'],
                                                  all_statuses):
                    styled_key = style_status(TASK_CLASS_MAPPING[key], status)
                    if error:
                        for line in error.splitlines():
                            click.echo(f'[{styled_key}] {line}')
                    else:
                        click.echo(f'[{styled_key}] {message} ... ')



#!/usr/bin/env python
# Copyright 2020 Grid AI Inc.
"""Entrypoint for the Grid CLI."""
import sys

import click
from click import ClickException

import grid.cli as cli
import grid.globals as env
from grid.metadata import __logo__
from grid.metadata import __version__
from grid.tracking import Segment
from grid.tracking import TrackingEvents
from grid.utilities import introspect_module


@click.group()
@click.option(
    '--debug',
    type=bool,
    help='Used for logging additional information for debugging purposes.',
    is_flag=True)
def main(debug: bool = True) -> None:
    """Grid CLI"""
    if debug:
        env.logger.info('Starting gridrunner in DEGUB mode.')

    env.DEBUG = debug


@main.command()
def version() -> None:
    """
    Prints CLI version to stdout
    """
    logo = click.style(__logo__, fg='green')
    click.echo(logo)

    version = f"""
                                Grid CLI ({__version__})
                               https://docs.grid.ai
    """
    click.echo(version)


#  Adds all CLI commands. Commands are introspected
#  from the cli module.
for command in introspect_module(cli):
    command: click.Command
    main: click.Group
    main.add_command(command)

if __name__ == '__main__':
    tracker = Segment()
    tracker.send_event(event=TrackingEvents.CLI_STARTED)
    try:
        main.main(prog_name="grid", standalone_mode=False)
        tracker.send_event(event=TrackingEvents.CLI_FINISHED)
    except ClickException as e:
        tracker.report_exception(e)
        e.show()
        sys.exit(e.exit_code)
    except Exception as e:
        tracker.report_exception(e)
        sys.exit(1)
    finally:
        tracker.flush()



from typing import List

import click

from grid import Grid


@click.command()
@click.argument('experiment', type=str, nargs=1)
@click.argument('metric', type=str, nargs=1)
@click.option('--n_lines',
              default=10,
              type=int,
              help='Number of metrics lines (from the end) to return.')
def metrics(experiment: str, metric: str, n_lines: int) -> None:
    """Gets metric values from an experiment."""
    # Instantiate client.
    client = Grid()

    # Call client.
    client.experiment_metrics(experiment_id=experiment,
                              metric=metric,
                              n_lines=n_lines)



import click

from grid.client import Grid
import grid.globals as env
from grid.types import ObservableType
from grid.utilities import is_experiment


@click.command()
@click.argument('run_or_experiment', type=str, nargs=1)
@click.argument('page', type=str, nargs=1, required=False)
def view(run_or_experiment: str, page: str) -> None:
    """Grid view shows we web UI page for your runs and experiments."""
    # Fetch URL from globals.
    url = env.GRID_URL.replace('/graphql', '#')

    # Instantiate Grid client.
    client = Grid()

    # Figure out which object is requested
    # so we can construct path.
    base_path = 'view'
    qualifier_path = ''
    if is_experiment(run_or_experiment):
        qualifier_path = 'experiment'
        kind = ObservableType.EXPERIMENT
    else:
        qualifier_path = 'run'
        kind = ObservableType.RUN

    # Combine all strings into a single URL.
    launch_url = None
    if page:
        launch_url = '/'.join(
            [url, base_path, qualifier_path, run_or_experiment, page])
    else:
        launch_url = '/'.join(
            [url, base_path, qualifier_path, run_or_experiment])

    # If the page requested is Tensorboard
    # get the URLs for those services from the backend and
    # open those specific pages in the browser.
    #
    # We'll use Tensorboard filtering to construct the right URL.
    error_message = ''
    tensorboard_url = None
    if page == 'tensorboard':

        # If the user has asked for aexperiment, extract
        # the run name from the identn ifier so we an fetch the
        # Tensorboard instance for that.
        run_name = run_or_experiment
        if kind == ObservableType.EXPERIMENT:
            run_name = '-'.join(run_or_experiment.split('-')[:-1])

        # Always get the status for a given Run because
        # that's where we store the resource URLs.
        observable = client.status(kind=ObservableType.RUN,
                                   identifiers=[run_name])

        # Finds the run result.
        run_data = None
        for run in observable['getRuns']:
            if run['name'] == run_name:
                run_data = run
                break

        # Construct the URLs for both services.
        if run_data:
            resource_urls = run_data.get('resourceUrls')
            if resource_urls:
                tensorboard_url = run_data['resourceUrls'].get('tensorboard')
                if tensorboard_url:
                    tensorboard_url = f"{tensorboard_url}#scalars&regexInput={run_or_experiment}"

            if not tensorboard_url:
                error_message = "Tensorboard isn't ready yet."

            launch_url = tensorboard_url

    # If we could not find the requested URL, raise an error.
    if not launch_url:
        raise click.ClickException(
            f'Could not view page {page} for {run_or_experiment}. {error_message}'
        )

    # Open browser.
    click.echo()
    click.echo(f'Opening URL: {launch_url}')
    click.echo()

    click.launch(launch_url)



import click

from grid import Grid
from grid.utilities import is_experiment


@click.command()
@click.argument('run_or_experiment', type=str, required=True, nargs=1)
def delete(run_or_experiment: str) -> None:
    """Deletes running run or experiment."""
    client = Grid()
    confirmed_is_experiment = is_experiment(run_or_experiment)

    # Confirm that the user really wants to do this.
    object_str = 'Experiment' if confirmed_is_experiment else 'Run'
    warning_str = click.style('WARNING!', fg='red')
    message = f"""

    {warning_str}

    Your are about to delete the {object_str} ({run_or_experiment}).
    This will delete all the associated artifacts, logs, and metadata.

    Are you sure you want to do this?

   """
    click.confirm(message, abort=True)

    if confirmed_is_experiment:
        client.delete(experiment_id=run_or_experiment)
    else:
        client.delete(run_id=run_or_experiment)



from typing import Optional

import click

from grid import Grid
from grid.utilities import check_is_experiment


@click.command()
@click.argument('experiment', type=str, nargs=1, callback=check_is_experiment)
@click.option('--n_lines',
              default=10,
              type=int,
              help='Number of log lines (from the end) to return.')
@click.option('--page',
              type=int,
              help='Which page to fetch from archived logs')
@click.option('--max_lines',
              type=int,
              help='Maximum number of lines to print in terminal')
@click.option(
    '--use_pager',
    is_flag=True,
    default=False,
    help='If the log output should be displayed as a pager scrollable')
def logs(experiment: str, n_lines: int, page: Optional[int],
         max_lines: Optional[int], use_pager: bool) -> None:
    """Gets logs from an experiment."""
    # Instantiate client.
    client = Grid()

    # Call client.
    client.experiment_logs(experiment_id=experiment,
                           n_lines=n_lines,
                           page=page,
                           max_lines=max_lines,
                           use_pager=use_pager)



import random
import subprocess
import sys
from typing import Dict, Optional

import click
from coolname import generate_slug
import yaml

from grid import Grid
from grid.cli.grid_credentials import print_creds_table
from grid.cli.grid_train import _check_run_name_is_valid
from grid.cli.utilities import read_config
from grid.cli.utilities import validate_config
from grid.cli.utilities import validate_disk_size_callback
import grid.globals as env
from grid.tracking import Segment
from grid.tracking import TrackingEvents
from grid.types import ObservableType
from grid.utilities import check_description_isnt_too_long


def _print_submission_message(instance_type: str, provider: str,
                              credential_id: str, name: str,
                              datastore_name: str, datastore_version: str,
                              datastore_mount_dir: str):
    """Prints train submission message to the client."""
    message = f"""
        Interactive node created!

        `grid status` to list all runs and interactive nodes.
        `grid status {name}` to see the status fot this interactive node.

        ----------------------
        Submission summary
        ----------------------
        name:                    {name}
        instance_type:           {instance_type}
        cloud_provider:          {provider}
        cloud_credentials:       {credential_id}
        datastore_name:          {datastore_name}
        datastore_version:       {datastore_version}
        datastore_mount_dir:     {datastore_mount_dir}
        """

    click.echo(message)


def _generate_default_interactive_config(
        client: Grid,
        credential_id: str,
        instance_type: str,
        datastore_name: Optional[str] = None,
        datastore_version: Optional[str] = None,
        datastore_mount_dir: Optional[str] = None,
        disk_size: Optional[int] = None):
    """
    Generates a new default config file for user if user hasn't
    submitted one.
    """
    _default_credential = None
    creds = client.get_credentials()['getUserCredentials']
    if credential_id:
        if creds:
            #  Check if the credential passed is can
            #  be used by user. We'll raise an exception
            #  if user can't use such credential.
            for cred in creds:
                if cred['credentialId'] == credential_id:
                    _default_credential = cred
                    break

        if not _default_credential:
            raise click.ClickException(
                f'Credential ID {credential_id} does not exist. ' +
                'Use grid credentials to see available credential IDs.')
    else:
        _url = client.url.replace('/graphql', '')
        no_credentials_message = f"""
        No cloud credentials available! Visit {_url}/#/settings to
        add new cloud credentials.
        """
        if not creds:
            raise click.ClickException(no_credentials_message)

        if len(creds) == 1:
            cred = creds[0]
            _default_credential = cred

        elif len(creds) > 1:
            for credential in creds:
                if credential['defaultCredential']:
                    click.echo(
                        f"Using default cloud credentials {credential['credentialId']} "
                        f"to run on {credential['provider'].upper()}.")
                    _default_credential = credential
                    break

        #  If no credentials are available, raise an exception.
        if not _default_credential:
            m = """
            Detected multiple credentials. Which would you like to use?

                grid interactive --grid_credential [CREDENTIAL_ID]

            """
            print_creds_table(creds)
            raise click.ClickException(m)

    #  Relevant defaults for user.
    #  TODO: Pull from user profile when we have
    #  that information collected.
    defaults = {'region': 'us-east-1'}

    _grid_config = {
        'compute': {
            'provider': {
                'vendor': _default_credential['provider'],
                'credentials': _default_credential['credentialId'],
                'region': defaults['region']
            },
            'train': {
                'instance': instance_type,
                'disk_size': disk_size,
                #  These parameters are required for the
                #  training config. The max_nodes need to be a value
                #  other than 0, otherwise subsequent train operations
                #  will not work.
                'nodes': 0,  # Start with 0 compute nodes.
                'max_nodes': 10,  # Scale at most to 100 nodes.
                'gpus': 0,  # Required, but ignored.
                'datastore_name': datastore_name,
                'datastore_version': datastore_version,
                'datastore_mount_dir': datastore_mount_dir,
                'scale_down_seconds':
                30 * 60,  # How many seconds before scaling down?
            }
        }
    }

    #  Print debug message if in debug mode.
    if env.DEBUG:
        click.echo('Grid Config used:')
        click.echo(_grid_config)

    return _grid_config


@click.group(invoke_without_command=True)
@click.pass_context
def interactive(ctx) -> None:
    """Displays interactive nodes for given project."""
    client = Grid()

    if ctx.invoked_subcommand is None:
        # Get the status of the interactive observables.
        kind = ObservableType.INTERACTIVE
        client.status(kind=kind, follow=False)


@interactive.command()
@click.option('--grid_config',
              '--g_config',
              'config',
              type=click.File('r'),
              required=False,
              callback=read_config,
              help='Path to Grid config YML')
@click.option('--grid_name',
              '--g_name',
              'name',
              type=str,
              required=False,
              help='Name for this run',
              callback=_check_run_name_is_valid)
@click.option('--grid_description',
              '--g_description',
              'description',
              type=str,
              required=False,
              help='Interactive node description; useful for note-keeping',
              callback=check_description_isnt_too_long)
@click.option('--grid_credential',
              '--g_credential',
              'credential',
              type=str,
              required=False,
              help='Cloud to run on')
@click.option('--grid_instance_type',
              '--g_instance_type',
              'instance_type',
              type=str,
              default='g3.8xlarge',
              help='Instance type to start training session in')
@click.option('--grid_disk_size',
              '--g_disk_size',
              'disk_size',
              type=int,
              required=False,
              default=200,
              callback=validate_disk_size_callback,
              help='The disk size in GB to allocate to interactive node')
@click.option('--grid_datastore_name',
              '--g_datastore_name',
              'datastore_name',
              type=str,
              required=False,
              help='Datastore name to be mounted in interactive node')
@click.option('--grid_datastore_version',
              '--g_datastore_version',
              'datastore_version',
              type=str,
              required=False,
              help='Datastore version to be mounted in interactive node')
@click.option(
    '--grid_datastore_mount_dir',
    '--g_datastore_mount_dir',
    'datastore_mount_dir',
    type=str,
    required=False,
    default='/home/jovyan/datastore',
    help=
    'Absolute path to mount Datastore in interactive node. (default to ~/datastore)'
)
def create(config: Optional[Dict], name: Optional[str],
           description: Optional[str], credential: Optional[str],
           instance_type: str, datastore_name: Optional[str],
           datastore_version: Optional[str],
           datastore_mount_dir: Optional[str],
           disk_size: Optional[int]) -> None:
    """Creates an interactive node."""
    client = Grid()

    tracker = Segment()
    tracker.send_event(event=TrackingEvents.INTERACTIVE_NODE_CREATED,
                       properties={'user_input': " ".join(sys.argv)})

    # make a fun random name when user does not pass in a name
    if name is None:
        name = f'{generate_slug(2)}-{random.randint(0, 1000)}'
        click.echo(
            f'No --grid_name passed, naming your interactive node: {name}')

    #  If the user has not passed a grid config file,
    #  then generate one with a set of default options.
    #  We'll add a default instance and the user's
    #  default credentials.
    if not config:
        _grid_config = _generate_default_interactive_config(
            client=client,
            datastore_name=datastore_name,
            datastore_version=datastore_version,
            datastore_mount_dir=datastore_mount_dir,
            credential_id=credential,
            instance_type=instance_type,
            disk_size=disk_size)

    else:
        _grid_config = config

    validate_config(cfg=_grid_config)

    client.validate_datastore_version(grid_config=_grid_config)
    datastore_version = _grid_config["compute"]["train"].get(
        "datastore_version")

    _grid_config_yaml = yaml.dump(_grid_config)

    #  Send to client.
    client.create_interactive_node(config=_grid_config_yaml,
                                   name=name,
                                   description=description)

    #  Shows a friendly submission message to the
    #  user.
    _print_submission_message(
        instance_type=_grid_config['compute']['train']['instance'],
        provider=_grid_config['compute']['provider']['vendor'],
        credential_id=_grid_config['compute']['provider']['credentials'],
        datastore_name=datastore_name,
        datastore_version=datastore_version,
        datastore_mount_dir=datastore_mount_dir,
        name=name)


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
def pause(interactive_node: str) -> None:
    """Pauses an interactive node."""
    client = Grid()
    client.pause_interactive_node(interactive_node_name=interactive_node)


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
def resume(interactive_node: str) -> None:
    """Resumes an interactive node."""
    client = Grid()
    client.resume_interactive_node(interactive_node_name=interactive_node)


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
def delete(interactive_node: str) -> None:
    """Deletes an interactive node."""
    client = Grid()
    client.delete_interactive_node(interactive_node_name=interactive_node)


@interactive.command()
def sync_ssh_config():
    """
    Sync interactive node's ssh config to the local ssh config.

    It manages a section within user's ssh config file
    for all interactive nodes shh config details

    Afterwards you can use systems ssh & related utilities
    (sshfs, rsync, ansible, whatever) with interactive nodes
    directly

    The default file is ~/.ssh/config and  can be changed via
    envvar GRID_SSH_CONFIG
    """
    client = Grid()
    nodes = client.sync_ssh_config()
    click.echo(f"Sync config for {len(nodes)} interactive nodes")


@interactive.command()
@click.argument('node_name', type=str)
def ssh(node_name):
    """
    SSH to interactive nodes.

    Usage: grid interactive ssh <node name>

    This command sync your ~/.ssh/config entries which are known to grid
    and executes ssh <node name>
    If you'd like the full power of ssh, you can use any ssh client and
    do `ssh <node_name>`. This command is stripped down version of it.

    See Also:
        grid interactive sync-ssh-config --help
    \f
    """
    client = Grid()
    client.sync_ssh_config()
    subprocess.check_call(['ssh', node_name])


@interactive.command()
@click.argument('interactive_node', type=str, nargs=1)
@click.argument('mount_dir', type=str, nargs=1)
def mount(interactive_node, mount_dir):
    """
    Mount interactive node directory to local.

    To mount a filesystem use:
    ixNode:[dir] mountpoint

    Examples:
        # Mounts the home directory on the interactive node in dir data
        grid interactive mount bluberry-122 ./data

        # mounts ~/data directory on the interactive node to ./data
        grid interactive mount bluberry-122:~/data ./data

   To unmount it:
      fusermount3 -u mountpoint   # Linux
      umount mountpoint           # OS X, FreeBSD

    Under the hood this is just passing data to sshfs after syncing grid's interactive,
    i.e. this command is dumbed down sshfs

    See Also:
        grid interactive sync-ssh-config --help
    \f
    """
    if ':' not in interactive_node:
        interactive_node += ":/home/jovyan"
    client = Grid()
    client.sync_ssh_config()
    subprocess.check_call(['sshfs', interactive_node, mount_dir])



from typing import List, Optional

import click

from grid import Grid
from grid.utilities import check_is_experiment


@click.command()
@click.option('--download_dir',
              type=click.Path(exists=False, file_okay=False, dir_okay=True),
              required=False,
              default='./grid_artifacts',
              help='Download directory that will host all artifact files')
@click.argument('experiments',
                type=str,
                required=True,
                nargs=-1,
                callback=check_is_experiment)
def artifacts(experiments: List[str],
              download_dir: Optional[str] = None) -> None:
    """Downloads artifacts for a given experiment or set of experiments."""
    client = Grid()

    for experiment in experiments:
        client.download_experiment_artifacts(experiment_id=experiment,
                                             download_dir=download_dir)



import click

from grid import Grid
from grid.utilities import is_experiment


@click.command()
@click.argument('runs_or_experiments', type=str, required=True, nargs=-1)
def cancel(runs_or_experiments: str) -> None:
    """Cancels running runs or experiments"""
    client = Grid()

    for run_or_exp in runs_or_experiments:
        #  Cancel Experiment
        if is_experiment(run_or_exp):
            client.cancel(experiment_id=run_or_exp)

        #  Cancel Run
        else:
            client.cancel(run_name=run_or_exp)
